import logging


class MaxLevelFilter(logging.Filter):
    def __init__(self, max_level: str) -> None:
        # max_level é um argumento adicional que coloquei
        # max_level deve receber o nome do level, mas quero o número
        super().__init__()

        # self.max_level terá o número do level
        self.max_level = logging.getLevelNamesMapping().get(max_level.upper(), 50)

    def filter(self, record: logging.LogRecord) -> bool:
        # record é o LogRecord que eu disse antes
        # record.levelno é o número do level do log
        # se o número do level do log for menor ou igual ao max_level que
        # definimos no filter o log passa.
        # INFO 20 só aceitará logs INFO e DEBUG.
        return record.levelno <= self.max_level


from collections.abc import Callable
from os import getenv
from pathlib import Path
from typing import Literal

from dotenv import load_dotenv

load_dotenv()

type LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
ALLOWED_LEVELS: set[LogLevel] = {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}


def define_setting[T](value: T, validator: Callable[[T], T] | None = None) -> T:
    if validator is not None:
        return validator(value)
    return value


def validate_path_dir(path: Path) -> Path:
    if not path.is_dir():
        raise NotADirectoryError(path)
    return path


def validate_path_file(path: Path) -> Path:
    if not path.is_file():
        raise FileNotFoundError(path)
    return path


def validate_level(level: str) -> LogLevel:
    if level not in ALLOWED_LEVELS:
        msg = f"Level {level!r} is not allowed. Use one of these: {ALLOWED_LEVELS}"
        raise ValueError(msg)
    return level


root_dir = Path(".").resolve()
logs_dir = root_dir / getenv("LOGS_DIR", "logs")
logging_config_json = root_dir / getenv("LOGGING_CONFIG_JSON", "logging.conf.json")

setup_logger_name = getenv("SETUP_LOGGER_NAME", "WARNING")
setup_logger_level = getenv("SETUP_LOGGER_LEVEL", "WARNING")

default_logger_level = getenv("DEFAULT_LOGGER_LEVEL", "WARNING")


def validate() -> None:
    global \
        root_dir, \
        logs_dir, \
        logging_config_json, \
        setup_logger_name, \
        setup_logger_level, \
        default_logger_level

    root_dir = define_setting(root_dir, validator=validate_path_dir)
    logs_dir = logs_dir / getenv("LOGS_DIR", "logs")
    logging_config_json = define_setting(
        root_dir / getenv("LOGGING_CONFIG_JSON", "logging.conf.json"),
        validator=validate_path_file,
    )

    setup_logger_name = getenv("SETUP_LOGGER_NAME", "config_setup")

    default_logger_level = define_setting(
        getenv("SETUP_LOGGER_LEVEL", "WARNING"), validator=validate_level
    )

    default_logger_level = define_setting(
        getenv("DEFAULT_LOGGER_LEVEL", "WARNING"), validator=validate_level
    )


def change_settings(
    new_root_dir: Path | None = None,
    new_logs_dir: Path | None = None,
    new_logging_config_json: Path | None = None,
    new_setup_logger_name: str | None = None,
    new_setup_logger_level: LogLevel | None = None,
    new_default_logger_level: LogLevel | None = None,
) -> None:
    global \
        root_dir, \
        logs_dir, \
        logging_config_json, \
        setup_logger_name, \
        setup_logger_level, \
        default_logger_level

    if new_root_dir:
        root_dir = define_setting(new_root_dir, validator=validate_path_dir)

    if new_logs_dir:
        logs_dir = define_setting(new_logs_dir, validator=validate_path_dir)

    if new_logging_config_json:
        logging_config_json = define_setting(
            new_logging_config_json,
            validator=validate_path_file,
        )

    if new_setup_logger_name:
        setup_logger_name = new_setup_logger_name

    if new_setup_logger_level:
        setup_logger_level = define_setting(
            new_setup_logger_level, validator=validate_level
        )

    if new_default_logger_level:
        default_logger_level = define_setting(
            new_default_logger_level, validator=validate_level
        )



import atexit
import json
import logging
from logging.config import dictConfig
from logging.handlers import QueueHandler, QueueListener

from logery.settings import (
    LogLevel,
    default_logger_level,
    logging_config_json,
    logs_dir,
    setup_logger_level,
    setup_logger_name,
    validate,
    validate_level,
)

_setup_logging_done: bool = False
_default_queue_listener: QueueListener | None = None

_logger = logging.getLogger(setup_logger_name)
_logger.setLevel(setup_logger_level)


def _setup_logging() -> None:
    global _setup_logging_done, _default_queue_listener

    if _setup_logging_done:
        _logger.debug("logging already configured, doing nothing for now")
        return

    validate()

    if not logging_config_json.is_file():
        msg = f"Logging config file does not exist: {logging_config_json}"
        raise FileNotFoundError(msg)

    if not logs_dir.is_dir():
        logs_dir.mkdir(parents=True, exist_ok=True)
        _logger.debug("Logs directory created: %s", logs_dir)

    with logging_config_json.open("r", encoding="utf-8") as file:
        logging_config = json.load(file)
        _logger.debug("JSON config file loaded: %s", logging_config_json)

    dictConfig(logging_config)

    queue_handlers = [
        handler
        for handler in logging.getLogger().handlers
        if isinstance(handler, QueueHandler)
    ]

    queue_handlers_count = len(queue_handlers)
    _logger.debug("QueueHandlers found: %d", queue_handlers_count)

    if queue_handlers_count > 1:
        msg = "This function does not allow more than one QueueHandler"
        raise RuntimeError(msg)

    if queue_handlers_count > 0:
        queue_handler = queue_handlers[0]
        _logger.debug("Found QueueHandler with name: '%s'", queue_handler.name)

        if queue_handler:
            _default_queue_listener = queue_handler.listener

            if _default_queue_listener is not None:
                _default_queue_listener.start()
                atexit.register(_stop_queue_listener)

                _logger.debug(
                    "QueueListener from QueueHandler '%s' started", queue_handler.name
                )

                _logger.debug(
                    "Function '%s' registered with atexit",
                    _stop_queue_listener.__name__,
                )

    _setup_logging_done = True


def _stop_queue_listener() -> None:
    if _default_queue_listener is None:
        return

    _logger.debug("Default listener will stop now, 👋 bye...")
    _default_queue_listener.stop()


def get_logger(name: str = "", level: LogLevel | None = None) -> logging.Logger:
    if not _setup_logging_done:
        _setup_logging()
        _logger.debug("'_setup_logging' used to configure Python logging.")

    logger = logging.getLogger(name)

    if level is not None:
        validate_level(level)
        _logger.debug(
            f"Level {level!r} used by 'get_logger' to configure {name!r} logger."
        )
        logger.setLevel(level)
    else:
        env_level = default_logger_level
        _logger.debug(
            f"Level {env_level!r} used by 'ENV' to configure {name!r} logger."
        )
        logger.setLevel(env_level)

    return logger



from setuptools import setup, find_packages


setup(
    name='logery',
    version='1.1.0',
    packages=find_packages(),
    install_requires=[
        'rich',
        'python-dotenv'
    ],
    description='Logs personalizados',
    author='Miguel Tenório',
    author_email='deepydev42@gmail.com',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.12',
)



# Aula 11
# Como criar um Formatter JSON do zero para Python Logging - Aula 11
# https://youtu.be/jX4Ai-ZWkj4
#
# Playlist:
# https://www.youtube.com/playlist?list=PLbIBj8vQhvm28qR-yvWP3JELGelWxsxaI
#
# Artigo:
#
# https://www.otaviomiranda.com.br/2025/logging-no-python-pare-de-usar-print-no-lugar-errado/#criando-um-json-log-formatter

import json
import logging
from datetime import datetime
from typing import Any, override
from zoneinfo import ZoneInfo

# Constantes de Configuração
# Define o fuso horário para os logs, garantindo consistência independente do servidor.
# https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
TZ_IDENTIFIER = "America/Sao_Paulo"
TZ = ZoneInfo(TZ_IDENTIFIER)

LOG_RECORD_KEYS = [
    "name",
    "msg",
    "args",
    "levelname",
    "levelno",
    "pathname",
    "filename",
    "module",
    "exc_info",
    "exc_text",
    "stack_info",
    "lineno",
    "funcName",
    "created",
    "msecs",
    "relativeCreated",
    "thread",
    "threadName",
    "processName",
    "process",
    "taskName",
    "message",
]


class JSONLogFormatter(logging.Formatter):
    def __init__(
        self,
        include_keys: list[str] | None = None,
        datefmt: str = "%Y-%m-%dT%H:%M:%S%z",
    ) -> None:
        super().__init__()
        self.include_keys = (
            include_keys if include_keys is not None else LOG_RECORD_KEYS
        )
        self.datefmt = datefmt

    @override
    def format(self, record: logging.LogRecord) -> str:
        dict_record: dict[str, Any] = {
            key: getattr(record, key)
            for key in self.include_keys
            if key in LOG_RECORD_KEYS and getattr(record, key, None) is not None
        }

        if "created" in dict_record:
            # Sobrescrevi o método `formatTime` para retornar um datetime
            # ao invés de `struct_time` que é o padrão. Assim consigo trabalhar
            # com timezone.
            dict_record["created"] = self.formatTime(record, self.datefmt)

        if "message" in self.include_keys:
            dict_record["message"] = record.getMessage()

        if "exc_info" in dict_record and record.exc_info:
            # `exc_info` traz informações sobre exceções. Precisamos formatar
            # esse valor para uma string. Por sorte isso existe em `Formatter`.
            dict_record["exc_info"] = self.formatException(record.exc_info)

        if "stack_info" in dict_record and record.stack_info:
            # Aqui também precisamos formatar o valor do stack da exceção para str.
            dict_record["stack_info"] = self.formatStack(record.stack_info)

        # Caso utilize extras ao emitir o log
        # Ex.: logger.warning("Mensagem", extra={"contexto": "qualquer coisa"})
        # A chave "contexto" será adicionada ao log também
        for key, val in vars(record).items():
            if key in LOG_RECORD_KEYS:
                # Essas chaves nós tratamos antes
                continue

            if key not in self.include_keys:
                msg = f'Key {key!r} does not exist in "include_keys"'
                raise KeyError(msg)

            # Adicionamos a chave extra ao log
            dict_record[key] = val

        return json.dumps(dict_record, default=str)

    @override
    def formatTime(self, record: logging.LogRecord, datefmt: str | None = None) -> str:
        date = datetime.fromtimestamp(record.created, tz=TZ)

        if datefmt:
            return date.strftime(datefmt)

        return date.isoformat()
